#! /bin/bash
# A small utility to build fixture servers
# for non regression testing of codegen
shopt -s extglob

# function initColors() {
if [[ -z "${MONO}" ]]; then
    normal=$(tput sgr0)
    red=$(tput setaf 1)
    green=$(tput setaf 2)
    orange=$(tput setaf 3)
    bold=$(tput bold)
    blue=$(tput setaf 4)
else
    normal=""
    red=""
    green=""
    orange=""
    blue=""
    bold=""
fi
# }

function okcr() {
    ok "$*"; printf "\n"
}
function ok() {
    printf "${green}%s${normal}" "$*"
}
function errcr() {
    err "$*" ; printf "\n"
}
function err() {
    printf "${red}${bold}%s${normal}" "$*"
}
function successcr() {
    success "$*" ; printf "\n"
}
function success() {
    printf "${green}${bold}%s${normal}" "$*"
}
function warncr() {
    warn "$*" ; printf "\n"
}
function warn() {
    printf "${orange}%s${normal}" "$*"
}
function infocr() {
    info "$*" ; printf "\n"
}
function info() {
    printf "${blue}%s${normal}" "$*"
}

if [ ! -f `which swagger` ]; then
  echo "can't find swagger in the PATH"
  exit 1
fi

# NOTE(fredbi):
# The following ones fail generation:
# - existing-model.yml requires pregeneration (not supported yet by this script)
# - issue72: model works with --skip-validation. Invalid spec (duplicate operationID)
# - todolist.simple.yml: invalid default values put on purpose for UT (provided fixed version for testing)
# - fixture-basetypes.yaml: exhibits some edge case failures with discriminator (e.g. using a base type in tuple...)
# - fixture-polymorphism.yaml: idem
#
# The following ones requires some checks to be skipped:
# - azure-text-analyis.json: works with --skip-validation. Invalid specification with duplicate property in allOf construct (provided fixed version for testing)
# - swagger-gsma.json: idem
#
# The following ones used to fail and are ok:
# - todolist.discriminators.yml: works (not with expand)
# - bitbucket.json: works nows (not with expand: too many files for ordinary ulimit)
# - cloudbreak.json : now works
known_failed="@(\
existing-model.yml|\
issue72.json|\
todolist.simple.yml|\
fixture-basetypes.yaml|\
fixture-polymorphism.yaml|\
)"
# The following ones should fail validation, but produce correct generated code (at least it builds)
known_skip_validation="@(\
todolist.enums.yml|\
todolist.enums.flattened.json|\
todolist.models.yml|\
todolist.schemavalidation.yml|\
azure-text-analyis.json|\
swagger-gsma.json|\
fixture-844-variations.yaml|\
fixture-allOf.yaml|\
fixture-errors.yaml|\
fixture-itching.yaml|\
fixture-tuple.yaml|\
fixture-simple-tuple.yaml|\
)"

# A list of known client build failures
known_client_failure="@(\
todolist.arrayform.yml|\
todolist.arrayquery.yml|\
todolist.url.basepath.yml|\
todolist.url.simple.yml|\
swagger-codegen-tests.json|\
fixture-1414.json|\
fixture-909-3.yaml|\
fixture-909-4.yaml|\
fixture-909-5.yaml|\
fixture-909-6.yaml|\
gentest2.yaml|\
gentest3.yaml|\
gentest.yaml|\
fixture-1437-4.yaml|\
fixture-1392-2.yaml|\
fixture-1392-3.yaml|\
)"

# A list of known fixtures not supporting expand mode (not including the discriminator case).
# Normally, this is because of duplicate names constructed during codegen of anonymous structures.
# This should be solved with proper analysis of names before codegen.
known_expand_failure="@(\
todolist.enums.flattened.json|\
fixture-1479.yaml|\
bitbucket.json|\
)"

# A list of known failures with minimal spec flattening. For those special cases, a full
# flattening is required
known_failed_minimal="@(\
fixture-1767.yaml|\
)"

# this collects all available yml/yaml/json fixtures in the following directories
fixtureDirs=(\
"fixtures/codegen" \
"fixtures/bugs/909" \
"fixtures/bugs/1437" \
"fixtures/bugs/1314" \
"fixtures/bugs/1062/eve-online-esi.json" \
"fixtures/bugs/1392" \
"fixtures/bugs/1277" \
"fixtures/bugs/1536" \
"fixtures/bugs/1487" \
"fixtures/bugs/1571" \
"fixtures/bugs/957" \
"fixtures/bugs/1614" \
"fixtures/bugs/931" \
"fixtures/bugs/1683" \
"fixtures/bugs/1796" \
"fixtures/bugs/1839" \
"fixtures/bugs/1719" \
"fixtures/bugs/1859" \
"fixtures/bugs/1490" \
"fixtures/bugs/973" \
"fixtures/bugs/1020" \
"fixtures/bugs/1339" \
"fixtures/bugs/1893" \
"fixtures/bugs/1518" \
"fixtures/bugs/1993" \
"fixtures/bugs/1937" \
)

# there are several subspecs there: we just want a specific one
singleFixtures=(\
"fixtures/bugs/1621/fixture-1621.yaml" \
"fixtures/bugs/1774/def_api.yaml" \
"fixtures/bugs/1767/fixture-1767.yaml" \
"fixtures/bugs/1260/fixture-realiased-types.yaml" \
"fixtures/bugs/1260/test3-swagger.yaml fixtures/bugs/1260/test3-bis-swagger.yaml" \
"fixtures/bugs/1260/test3-ter-swagger.yaml fixtures/bugs/1260/test3-ter-swagger-flat.json" \
"fixtures/bugs/1851/fixture-1851.yaml" \
)

if [[ "$1" == "--circleci" ]] ; then
    # Coloured output not supported by default on CircleCI.
    # Forcing term to xterm is not enough: tput not available with minimalist env.
    MONO=1
    # In CI run, test builds with default option (minimal flatten)
    OPTS="--with-flatten=minimal"
    #export TERM=xterm
    #MONO=""
else
    # In manual run, enable coloured output
    MONO=""
    # In manual run, test the full range of available options for spec preprocessing
    OPTS="--with-flatten=full --with-flatten=minimal --with-flatten=expand"
fi

# A little chrome does not hurt...
# initColors

# All fixtures in ./fixtures/codegen + some others
cd $(git rev-parse --show-toplevel)

# build list of fixtures from dirs
for fixture in ${fixtureDirs[@]} ; do
  specdir=${specdir}" ${fixture}"
done
check_list=`for d in ${specdir}; do ls $d/*.yml;ls $d/*.json;ls $d/*.yaml;done 2>/dev/null`

# extend list of fixtures with extra single specs
for fixture in ${singleFixtures[@]} ; do
  check_list=${check_list}" ${fixture}"
done

list=( $check_list )
fixtures_count=${#list[@]}

# cleanup CI spec generation target
gendir=./tmp-gen
rm -rf ${gendir}

okcr "Running codegen for ${fixtures_count} specs"

for spec in ${check_list}; do
    testcase=${spec##*/}
    case ${testcase} in
    ${known_failed})
        warncr "[`date +%T`]${spec}: not tested against full build because of known issues."
        run="false"
        opts=""
        buildClient="false"
        noexpand="false"
        ;;
    ${known_skip_validation})
        infocr "[`date +%T`]${spec}: assumed spec is invalid but tested against build nonetheless..."
        run="true"
        opts="--skip-validation"
        buildClient="true"
        noexpand="false"
        ;;
    ${known_client_failure})
        warncr "[`date +%T`]${spec}: will not attempt to build the client because of known issues..."
        run="true"
        opts=""
        buildClient="false"
        noexpand="false"
        ;;
    ${known_expand_failure})
        warncr "[`date +%T`]${spec}: will not attempt to build with expand mode because of known issues..."
        run="true"
        opts=""
        buildClient="true"
        noexpand="true"
        ;;
    ${known_failed_minimal})
        warncr "[`date +%T`]${spec}: will not attempt to build with flatten minimal mode because of known issues..."
        run="true"
        opts=""
        buildClient="true"
        noflatten="true"
        ;;
    *)
        infocr "[`date +%T`]${spec}: assumed valid and tested against build."
        run="true"
        opts=""
        buildClient="true"
        noexpand="false"
        ;;
    esac

	if [[ ${run} == "true" ]]; then
        declare -i index=0
        for preprocessingOpts in ${OPTS} ; do
            index+=1
            infocr "Generation with options: ${preprocessingOpts} ${opts}"

            # Do not attempt to generate on expanded specs when there is a discriminator specified
	        grep -q discriminator ${spec}
	        discriminated=$?
	        if [[ ${discriminated} -eq 0 && ${preprocessingOpts} == "--with-flatten=expand" ]] ; then
	            warncr "Skipped ${testcase} with ${preprocessingOpts}: discriminator not supported in this mode"
	            continue
	        fi
          if [[ ${noexpand} != "true" && ${preprocessingOpts} == "--with-flatten=expand" ]] ; then
              continue
          fi
          if [[ ${noflatten} == "true" && ( ${preprocessingOpts} == "--with-flatten=minimal" || ${preprocessingOpts} == "" ) ]] ; then
            if [[ ${OPTS} == "--with-flatten=minimal" ]] ; then
              preprocessingOpts="--with-flatten=full"
            else
              continue
            fi
          fi

	        target=${gendir}/gen-${testcase%.*}${index}
	        target_client=${gendir}/gen-${testcase%.*}${index}"-client"

	        server_name="nrcodegen"
	        client_name="nrcodegen"
	        errlog=${gendir}/stderr.log

	        rm -rf ${target} ${target_client}
	        mkdir -p ${target} ${target_client}
	        rm -f ${errlog}

	        # Gen server
	        if ! swagger generate server --spec ${spec} --target ${target} --name=${server_name} --quiet ${opts} ${preprocessingOpts} 2>${errlog} ; then
	            errcr "Generation Failed"
	            if [[ -f ${errlog} ]] ; then errcr "$(cat ${errlog})" ; rm ${errlog};fi
	            exit 1
	        fi
	        okcr "Generation server OK"
	        rm -f ${errlog}
	        # Gen client

	        if ! swagger generate client --spec "${spec}" --target "${target_client}" --name "${client_name}" --quiet ${opts} "${preprocessingOpts}" 2>${errlog}; then
	            errcr "Generation Failed"
	            if [[ -f ${errlog} ]] ; then errcr `cat ${errlog}` ; rm ${errlog};fi
	            exit 1
	        fi
	        okcr "Generation client OK"

            #
            # Check build on generated artifacts
            #

	        # Build server
	        if ! (cd ${target}/cmd/${server_name}"-server"; go build) 2>${errlog} ; then
	            errcr "Server build Failed"
	            if [[ -f ${errlog} ]] ; then errcr `cat ${errlog}` ; rm ${errlog};fi
	            exit 1
	        fi
	        okcr "Server build OK"
	        # Build models if any produced
	        if [[ -d ${target}/models ]] ; then

	            if ! (cd ${target}/models ; go build) 2>${errlog}; then
	                errcr "Model build Failed"
	                if [[ -f ${errlog} ]] ; then errcr `cat ${errlog}` ; rm ${errlog};fi
	                exit 1
	            fi
	        fi
	        okcr "Models build OK"
	        # Build client
	        if [[ ${buildClient} == "false" ]] ; then
	            warncr "(no client built)"
	        else
	            if ! (cd ${target_client}/client ; go build) 2>${errlog} ; then
	                errcr "Client build Failed"
	                if [[ -f ${errlog} ]] ; then errcr `cat ${errlog}` ; rm ${errlog};fi
	                exit 1
	            fi
	            okcr "Client build OK"
	        fi
	        successcr " [All builds for ${spec}:  OK]"
	        rm -f ${errlog}
	        rm -rf ${target} ${target_client}
	    done
	fi
done
rm -rf ${gendir}
exit 0
